#include <X11/X.h>
#include <X11/Intrinsic.h>
#include <X11/Xatom.h>
#include <X11/StringDefs.h>
#include <X11/Xutil.h>
#include <X11/Xlib.h>
#include <X11/Shell.h>

#include <Xm/Xm.h>
#include <Xm/AtomMgr.h>
#include <Xm/Protocols.h>

extern Display *cx_display;
extern Screen  *cx_screen;
extern Window   cx_rootwindow;

extern Widget   cx_toplevel;
extern Window   cx_topwindow;

static Atom   WM_SAVE_YOURSELF = NULL;
static Atom   W_M_P = NULL;

static Display *cx_cl_display;

/* Session manager call to apps registered for SAVE_YOURSELF. */

static int    cx_SYsent;

/* This is the recursive call used only internally.
   The user call appears below it. */

static void 
cx_cl_sendSY(Window tree_top, int depth)
{
  Window      root_window, parent_window;
  Window     *child_wins;
  unsigned int  num_children;
  int       ii, pp, chx;

  Atom      actual_type;
  int       actual_format;
  unsigned long   nitems;
  unsigned long   bytes_leftover;
  unsigned char  *data = NULL;

  XClientMessageEvent XCME;

  Atom       *proto_list;
  int       proto_count = 0, matched = 0;

/* Get past the WM re-parenting but then stop. */

  if (depth > 3)
    return;

/* Get list of children.
   This call returns 0 on failure / 1 on success. */

  if (!XQueryTree(cx_cl_display, tree_top,
          &root_window,
          &parent_window,
          &child_wins,
          &num_children))
    return;

/* Test each (except our own) for WM_SAVE_YOURSELF protocol. */

  for (chx = 0; chx < num_children; chx++) {
    if (child_wins[chx] != cx_topwindow) {

      proto_count = 0;
      matched = 0;

      if ((XGetWindowProperty(cx_cl_display, child_wins[chx],
                  W_M_P,
                  0, 1000, False, AnyPropertyType,
                  &actual_type,
                  &actual_format,
                  &proto_count,
                  &bytes_leftover,
                  &data) == Success)
        && proto_count) {

        proto_list = (Atom *) data;

        for (pp = 0;
           !matched && (pp < proto_count);
           pp++)
          matched = (proto_list[pp] == WM_SAVE_YOURSELF);

        if (matched) {

/* ----------
         printf("\n %d %s",
                proto_count,
                XmGetAtomName(cx_cl_display, actual_type));

         for (pp = 0; pp < proto_count; pp++)
           printf("|%s=%d|",
                  XmGetAtomName(cx_cl_display, proto_list[pp]),
                  proto_list[pp]);
------------ */

/* Send the S_Y as a ClientMessage.

   Registered clients are expected to respond with
   an update (or touch) of their WM_COMMAND property,
   but where and how long should we wait for the response,
   or equivalently,
   how long to delay before getting what should be new?

   We could select for PropertyChange Notify events,
   check each off if and when it is updated.

   A simpler way is for us to DELETE each WM_COMMAND property,
   then anything we find later must have been an update,
   and finding nothing means an application has resigned.
   This DELETE could also have served as the Save Yourself event,
   but that is not what got written in the standard.

   THIS LOCAL PRACTICE INTERFERES WITH ICCCM ALLOWED PRACTICE,
   where a task with no change can do a zero length append. */

          XDeleteProperty(cx_cl_display,
                  child_wins[chx],
                  XA_WM_COMMAND);

/* This event works!
   Details NOT as described in Young or O'Reilly vols 1, 4, 6! */

          memset(&XCME, 0, sizeof(XCME));

          XCME.type = ClientMessage;

          XCME.display = cx_cl_display;
          XCME.window = child_wins[chx];

          XCME.message_type = W_M_P;
          XCME.format = 8 * sizeof(Atom);

          XCME.data.l[0] = WM_SAVE_YOURSELF;
          XCME.data.l[1] = CurrentTime;

/* Do not propagate to parent window if no client is interested.
   The parent would be root and we know they don't want it.
   Count them if successfully sent,
   no direct way to know if actually processed. */

          if (XSendEvent(cx_cl_display,
                   child_wins[chx],
                   False,
                   NULL,
                   &XCME))
            cx_SYsent++;
        }

        if (proto_list)
          XFree(proto_list);
      }

      else
        cx_cl_sendSY(child_wins[chx], depth + 1);
    }
  }
  XFlush(cx_cl_display);
  XFree(child_wins);
}

/* Call with NULL or empty "" to use existing local display connection;
   or with "host:serve_num" or "host:serve_num.screen_num" for a remote.
   If we open a remote connection we close it when done with it.
   See the XOpenDisplay manpage for all the name variations. */

int 
cx_sendSY(char *display_name)

{ static char   kws[] = "@(#)cx_sendSY.c	1.1 ARGOS 94/02/05 17:34:08";

  Window      tree_top;

/* Remote if we can.
   Note that DefaultScreen is the one we ask for in the Open. */

  if (display_name
    && strlen(display_name)) {
    if (cx_cl_display = XOpenDisplay(display_name))
      tree_top = DefaultScreenOfDisplay(cx_cl_display)->root;
    else
      return NULL;
  }

/* Local if we are open as we should be. */

  else if (cx_display) {
    cx_cl_display = cx_display;
    tree_top = cx_rootwindow;
  }

/* Called before initial calls. */

  else {
    printf("gui_init must be called before cx_sendSY\n");
    return NULL;
  }

/* Use the built-in Atom if it is defined else make one. */

  if (!W_M_P)
#ifdef XA_WM_PROTOCOLS
    W_M_P = XA_WM_PROTOCOLS;
#else
    W_M_P = XmInternAtom(cx_cl_display, "WM_PROTOCOLS", False);
#endif

  if (!WM_SAVE_YOURSELF)
    WM_SAVE_YOURSELF = XmInternAtom(cx_cl_display, "WM_SAVE_YOURSELF", False);

/* Start at top of tree. */

  cx_SYsent = 0;

  cx_cl_sendSY(tree_top, 0);

/* Close a remote connection. */

  if (cx_cl_display != cx_display)
    XCloseDisplay(cx_cl_display);

  return cx_SYsent;
}
