%{

/* uilex.l
   We resolve ambiguous strings according to what yacc wants,
   as indicated by want_token flag settings. */

#include <stdio.h>

#include <string.h>

#include "y.tab.h"
   
extern int want_token;
extern int yylineno;

#include "uitoc_heads.h"

int res_word (int);
int res_token, subnode;

static char block[20000];

extern YYSTYPE yylval;

%}

ltr		[a-zA-Z_]
dig		[0-9]
l_or_d		[a-zA-Z_0-9\/\.]
w_sp		[ \t\n]
other		.

%%

"/*"		skipcomm();

"#line"		poundline();

"%{"		{ copyout();
                  yylval.text = block;
                  return BLOCK; }

{w_sp}		;

END		{ return _END; }

{dig}+		{ sscanf (yytext, "%d", &yylval.ival);
		  return INTEGER; } }

"/"{l_or_d}+    { if ((want_token == SUBTREE)
		    && ((subnode = uisubtree(yytext)) > 0)) {
		      yylval.ival = subnode;
		      return SUBTREE; }

		  if ((res_token = res_word (TEXT)) == TEXT)
		      yylval.text = strdup(yytext);

		  return res_token; }

"_"{dig}{l_or_d}*       { if ((want_token == SIBLING)
		    && uisibling(yytext+1, 1)) {
		      yylval.text = strdup(yytext+1);
		      return SIBLING; }

		  else if ((want_token == NEW_NAME)
		    && uisibling(yytext+1, 0)) {
		      yylval.text = strdup(yytext+1);
		      return NEW_NAME; }

		  if ((res_token = res_word (TEXT)) == TEXT)
		      yylval.text = strdup(yytext+1);

		  return res_token; }

{l_or_d}+       { if ((want_token == SIBLING)
		    && uisibling(yytext, 1)) {
		      yylval.text = strdup(yytext);
		      return SIBLING; }

		  else if ((want_token == NEW_NAME)
		    && uisibling(yytext, 0)) {
		      yylval.text = strdup(yytext);
		      return NEW_NAME; }

		  if ((res_token = res_word (TEXT)) == TEXT)
		      yylval.text = strdup(yytext);

		  return res_token; }

{other}		{ return 0; }

%%

  skipcomm()
  { int ch;
    for ( ; ; ) {
      while (input() != '*') ;
      if (input() == '/')
          return;
      else
          unput(yytext[yyleng-1]);
    }
  }

  poundline()
  { int ich; char lbuf[20]; int ii = 0;
      while ((ich = input()) != '\n')
        if (isdigit(ich) && (ii < 20))
           lbuf[ii++] = ich;
      yylineno = atoi(lbuf);
  }

  copyout()
  { int ich, bct = 0;
    for ( ; ; ) {
      while ((ich = input()) != '}')
          block[bct++] = ich;
      if ((ich = input()) == '%') {
          block[bct++] = 0;
          return;
      }
      else {
          block[bct++] = '}';
          block[bct++] = ich;
      }
    }
  }

int res_word(int no_match_token)
{ static char   kws[] = "@(#)uilex.l	1.1 ARGOS 93/12/13 17:45:27";
  extern char yytext[];
  int c;

  static struct res_table {
    char   *rw_name;
    int     rw_yylex;
  } res_table[] = {
	"ACTION",	ACTION,
	"ACTION_MSG",	ACTION_MSG,
	"ALL",		ALL,
	"ALLOW",	ALLOW,
	"AND",		AND,
	"AT",		AT,
	"AUTO_OK",	AUTO_OK,
	"AUTO_SELECT",	AUTO_SELECT,
	"BODY_COLOR",	BODY_COLOR,
	"BOOLEAN",	BOOLEAN,
	"CHOICES",	CHOICES,
	"CHOOSE_ONE",	CHOOSE_ONE,
	"CLICK_ACTION",	CLICK_ACTION,
	"COLUMNS",	TAB,
	"CONFIRM",	CONFIRM,
	"CONFIRM_MSG",	CONFIRM_MSG,
	"COPY",		COPY,
	"CYCLE",	CYCLE,
	"DECODE",	DECODE,
	"DEFAULT",	DEFAULT,
	"DISALLOW",	DISALLOW,
	"DISCARD",	DISCARD,
	"DOUBLE",	DOUBLE,
	"DOES",		DOES,
	"EMPTY",	EMPTY,
	"ENABLE_CHOICE",ENABLE_CHOICE,
	"ENABLE_OK",	ENABLE_OK,
	"ENCODE",	ENCODE,
	"EVENT_TEST",	EVENT_TEST,
	"EVERY_TIME",	EVERY_TIME,
	"FILL_COLOR",	FILL_COLOR,
	"FLOAT",	FLOAT,
	"FOLLOWS",	FOLLOWS,
	"HELP",		HELP,
	"HIGHLIGHT",	HIGHLIGHT,
	"IMMEDIATE",	IMMEDIATE,
	"INCLUDE",	INCLUDE,
	"INITVAL",	INITVAL,
	"INT",		INT,
	"LABEL",	LABEL,
	"LABEL_COLOR",	LABEL_COLOR,
	"LAYOUT",	LAYOUT,
	"LEFT",		LEFT,
	"LOGICAL",	BOOLEAN,
	"MAX",		MAX,
	"MIN",		MIN,
	"MOUSE_PICK",	MOUSE_PICK,
	"NONE",		NONE,
	"NO_ECHO",	NO_ECHO,
	"NO_LABEL",	NO_LABEL,
	"OMIT",		OMIT,
	"ON_COLOR",	ON_COLOR,
	"ON_CR",	ON_CR,
	"ON_ENTRY",	ON_ENTRY,
	"ON_EXIT",	ON_EXIT,
	"OR",		OR,
	"OVERLAYS",	OVERLAYS,
	"PERMANENT",	PERMANENT,
	"PICK",		PICK,
	"PLACE",	PLACE,
	"POPUP",	POPUP,
	"READ_ONLY",	READ_ONLY,
	"REMAIN",	REMAIN,
	"RFILE",	RFILE,
	"RIGHT",	RIGHT,
	"ROWS",		VTAB,
	"SAME",		SAME,
	"SELECT",	SELECT,
	"SET",		SET,
	"SIZE",		SIZE,
	"STRING",	STRING,
	"STRUCTURED",	STRUCTURED,
	"SUMMARY",	SUMMARY,
	"SWITCH",	SWITCH,
	"TAB",		TAB,
	"TERMINAL",	TERMINAL,
	"TOGGLE",	BOOLEAN,
	"TOOLBAR",	TOOLBAR,
	"TO_LOWER",	TO_LOWER,
	"TO_UPPER",	TO_UPPER,
	"VADJUST",	VADJUST,
	"VTAB",		VTAB,
	"WIDTH",	WIDTH,
	"WINDOW",	WINDOW,
};

#define  T_END(v)      (v-1 + sizeof v / sizeof v[0])

   struct res_table *low = res_table,
                   *high = T_END(res_table),
                   *mid;

/* Search the reserved words table to see if this is one of them
   See 'man ascii' if unsure of order for any characters.
   If so, return the token value defined for this string.
   Otherwise return the user defined default token. */

   while (low <= high) {
      mid = low + (high-low)/2;

      c = strcmp(mid->rw_name, yytext);

      if (c == 0)
         return(mid->rw_yylex);
      else if (c < 0)
         low = mid+1;
      else
         high = mid-1;
   }

   return(no_match_token);
}

